## Import packages
__all__=[
    # 'analysis',
    # 'comparisons',
    'decomposition',
    'h5_handling',
    'helpers',
    # 'neural',
    # 'optic_flow',
    'pipelines',
    'point_tracking',
    'project',
    'rois',
    'spectral_analysis',
    'util',
    'visualization',
    'data_importing',
    # 'tests',
]

import torch  ## For some reason, it crashes if I don't import torch before other packages... RH 20221128


## Prepare cv2.imshow
import pkg_resources
installed_packages = {pkg.key for pkg in pkg_resources.working_set}
has_cv2_headless = 'opencv-contrib-python-headless' in installed_packages
has_cv2_normal = 'opencv-contrib-python' in installed_packages
if has_cv2_normal and not has_cv2_headless:
    run_cv2_imshow = True
elif has_cv2_headless and not has_cv2_normal:
    run_cv2_imshow = False
elif has_cv2_headless and has_cv2_normal:
    raise ValueError("Both opencv-contrib-python and opencv-contrib-python-headless are installed. Please uninstall one of them.")
elif not has_cv2_headless and not has_cv2_normal:
    run_cv2_imshow = False
    import warnings
    warnings.warn("Neither opencv-contrib-python nor opencv-contrib-python-headless are installed. Please install one of them to use cv2.imshow().")
else:
    raise ValueError("This should never happen. Please report this error to the developer.")

if run_cv2_imshow:
    def prepare_cv2_imshow():
        """
        This function is necessary because cv2.imshow() 
        can crash the kernel if called after importing 
        av and decord.
        RH 2022
        """
        import numpy as np
        import cv2
        test = np.zeros((1,300,400,3))
        for frame in test:
            cv2.putText(frame, "WELCOME TO FACE RHYTHM!", (10,50), cv2.FONT_HERSHEY_SIMPLEX, 0.7, (255,255,255), 2)
            cv2.putText(frame, "Prepping CV2", (10,100), cv2.FONT_HERSHEY_SIMPLEX, 0.7, (255,255,255), 2)
            cv2.putText(frame, "Calling this figure allows cv2.imshow ", (10,150), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255,255,255), 1)
            cv2.putText(frame, "to work without crashing if this function", (10,170), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255,255,255), 1)
            cv2.putText(frame, "is called before importing av and decord", (10,190), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255,255,255), 1)
            cv2.imshow('startup', frame)
            cv2.waitKey(1000)
        cv2.destroyWindow('startup')
    prepare_cv2_imshow()


for pkg in __all__:
    exec('from . import ' + pkg)


__version__ = '0.2.4'


####### data_importing #######

## Dataset_videos
from face_rhythm.data_importing import Dataset_videos
class test_Dataset_videos:
    def __init__(self):
        pass

    ## Test if type checking works
    def test_type_checking(self):
        import pytest
        
        ## test if paths_videos is a list of strings or string
        conditions_error = [
            1,
            [1],
            [1, 2],
            ['test', 2],
            [1, 'test'],
        ]
        for condition in conditions_error:
            with pytest.raises(AssertionError):
                Dataset_videos(paths_videos=condition)

        ## test if paths_videos exist
        conditions_error = [
            ['test'],
            ['test', 'test'],
        ]
        for condition in conditions_error:
            with pytest.raises(AssertionError):
                Dataset_videos(paths_videos=condition)



import sys
from pathlib import Path
import re

def test_python_version(
    look_for_verion_in_environmentYaml=True, 
    filename_environmentYaml='environment.yml',
    fallback_version=3.11,
    verbose=1,
):
    """
    Test python version.
    Either use the version specified in environment.yaml file or
     falls back to the version specified in fallback_version.
    RH 2022

    Args:
        look_for_verion_in_environmentYaml (bool):
            If True, look for the version in environment.yaml file.
            If False, use the version specified in fallback_version.
        filename_environmentYaml (str):
            The name of the environment.yaml file.
        fallback_version (float):
            The version to use if look_for_verion_in_environmentYaml is False.
        verbose (int):
            If 0, do not print anything.
            If 1, print warnings.
            If 2, print all below and info.

    Returns:
        success (bool): 
            True if the version is correct, False otherwise.
    """
    ## find path to repo and environment.yml files
    path_repo = str(Path(__file__).parent.parent.parent)

    ## check if environment.yml exists. If not, warn user and 
    ##  use fallback version
    if look_for_verion_in_environmentYaml:
        path_envYaml = str(Path(path_repo) / filename_environmentYaml)
        if not Path(path_envYaml).exists():
            print(f'FR Warning: {path_envYaml} does not exist. Using fallback version {fallback_version}') if verbose > 0 else None
            version_test = fallback_version
        else:
            ## Read environment.yml file. Find string that contains python version.
            ## Extract version number and convert to float
            with open(path_envYaml, 'r') as f:
                lines = f.readlines()
                for line in lines:
                    if re.search(' python=', line):
                        version_test = float(line.split('=')[1])
                        print(f'FR: Found environment specification for PYTHON VERSION {version_test} from line: {line} in {path_envYaml}') if verbose > 1 else None
                        break
    else:
        version_test = fallback_version

    ## Test if python version and subversion are correct
    version_system = float(f'{sys.version_info.major}.{sys.version_info.minor}')
    print(f'FR: PYTHON VERSION system: {version_system}') if verbose > 1 else None
    if version_test != version_system:
        print(f'FR Error: PYTHON VERSION {version_system} does not match specification: {version_test}. Please check your environment.')
        return False
    print(f'FR: PYTHON VERSION on system: {version_system} matches specification: {version_test}') if verbose > 1 else None
    # return True


def test_torch_cuda(verbose=1, device='cuda'):
    """
    Test to see if torch can do operations on GPU if CUDA is available.
    RH 2022

    Args:
        verbose (int):
            If 0, do not print anything.
            If 1, print warnings.
            If 2, print all below and info.
        device (str):
            The device to use. Default is 'cuda'.
    """
    import torch
    ## Check if CUDA is available
    if torch.cuda.is_available():
        print(f'FR: CUDA is available. Environment using PyTorch version: {torch.__version__}') if verbose > 1 else None
        arr = torch.rand(1000, 100).to(device)
        arr2 = torch.rand(100, 1000).to(device)
        arr3 = (arr @ arr2).mean().cpu().numpy()
        print(f'FR: Torch can do basic operations on GPU. Environment using PyTorch version: {torch.__version__}. Result of operations: {arr3}') if verbose > 1 else None

    else:
        print(f'FR Error: CUDA is not available. Environment using PyTorch version: {torch.__version__}') if verbose > 0 else None
        # return True


from pathlib import Path

import warnings
import pytest
import tempfile

from face_rhythm import helpers, util, h5_handling, pipelines


def test_pipeline_tracking_simple(dir_data_test):
    dir_temp = str(tempfile.TemporaryDirectory().name)
    dir_project = str(Path(dir_temp).resolve() / 'project')
    dir_inputs       = str(Path(dir_data_test).resolve() / 'inputs')
    dir_outputs_true = str(Path(dir_data_test).resolve() / 'outputs')

    defaults = util.get_default_parameters(
        directory_project=dir_project,
        directory_videos=str(Path(dir_inputs)),
        filename_videos_strMatch=r'demo.*',
        path_ROIs=str(Path(dir_inputs) / 'ROIs.h5'),
    )
    SEED = 0
    params_partial = {
            "steps": [
                "load_videos",
                "ROIs",
                "point_tracking",
                "VQT",
                "TCA",
            ],
            "project": {
                "overwrite_config": True,
                "update_project_paths": True,
                "initialize_visualization": False,
                "use_GPU": False,
                "random_seed": SEED,
                "verbose": 2,
            },
            "figure_saver": {
                "formats_save": [
                    "png"
                ],
                "kwargs_savefig": {
                    "bbox_inches": "tight",
                    "pad_inches": 0.1,
                    "transparent": True,
                    "dpi": 300,
                },
                "overwrite": True,
                "verbose": 2
            },
            "paths_videos": {
                "depth": 0,
            },
            "BufferedVideoReader": {
                "buffer_size": 1000,
                "prefetch": 1,
                "posthold": 1,
                "method_getitem": "by_video",
                "verbose": 1,
            },
            "Dataset_videos": {
                "contiguous": False,
                "frame_rate_clamp": None,
                "verbose": 2,
            },
            "ROIs": {
                "initialize":{
                    "select_mode": "file",
                    "verbose": 2,
                },
                "make_rois": {
                    "rois_points_idx": [
                        0,
                    ],
                    "point_spacing": 9,
                },
            },
            "PointTracker": {
                "contiguous": True,
                "params_optical_flow": {
                    "method": "lucas_kanade",
                    "mesh_rigidity": 0.025,
                    "mesh_n_neighbors": 8,
                    "relaxation": 0.0015,
                    "kwargs_method": {
                        "winSize": [
                            20,
                            20,
                        ],
                        "maxLevel": 2,
                        "criteria": [
                            3,
                            2,
                            0.03,
                        ],
                    },
                },
                "visualize_video": False,
                "params_visualization": {
                    "alpha": 0.2,
                    "point_sizes": 2,
                },
                "params_outlier_handling": {
                    "threshold_displacement": 150,
                    "framesHalted_before": 10,
                    "framesHalted_after": 10,
                },
                "verbose": 2,
            },
            "VQT_Analyzer": {
                "params_VQT": {
                    "Q_lowF": 4,
                    "Q_highF": 10,
                    "F_min": 1.0,
                    "F_max": 60,
                    "n_freq_bins": 36,
                    "win_size": 501,
                    "symmetry": 'center',
                    "taper_asymmetric": True,
                    "plot_pref": False,
                    "downsample_factor": 20,
                    "padding": "valid",
                    "batch_size": 10,
                    "return_complex": False,
                    "progressBar": True,
                },
                "normalization_factor": 0.95,
                "spectrogram_exponent": 1.0,
                "one_over_f_exponent": 0.5,
                "verbose": 2
            },
            "TCA": {
                "verbose": 2,
                "rearrange_data": {
                    "names_dims_array": [
                        "xy",
                        "points",
                        "frequency",
                        "time",
                    ],
                    "names_dims_concat_array": [
                        [
                            "xy",
                            "points",
                        ]
                    ],
                    "concat_complexDim": False,
                    "name_dim_concat_complexDim": "time",
                    "name_dim_dictElements": "session",
                    "method_handling_dictElements": "separate",
                    "name_dim_concat_dictElements": "time",
                    "idx_windows": None,
                    "name_dim_array_window": "time",
                },
                "fit": {
                    "method": "CP_NN_HALS",
                    "params_method": {
                        "rank": 10,
                        "n_iter_max": 200,
                        "init": "random",
                        "svd": "truncated_svd",
                        "tol": 1e-09,
                        "random_state": SEED,
                        "verbose": True,
                    },
                    "verbose": 2,
                },
                "rearrange_factors": {
                    "undo_concat_complexDim": False,
                    "undo_concat_dictElements": False,
                },
            },
        }
    params = helpers.prepare_params(
        params=params_partial, 
        defaults=defaults,
        error_on_missing_keys=False,
    )
    results = pipelines.pipeline_basic(params)

    ## Check run_data equality
    print(f"Checking run_data equality")
    paths_rundata_true = helpers.find_paths(
        dir_outer=dir_outputs_true,
        reMatch=r'.*',
    )
    paths_rundata_relative = [str(Path(p).relative_to(dir_outputs_true)) for p in paths_rundata_true]
    paths_rundata_test = [str(Path(dir_project) / Path(p)) for p in paths_rundata_relative]

    def load_file(path):
        if Path(path).suffix == '.pkl':
            return helpers.pickle_load(path)
        elif Path(path).suffix == '.json':
            return helpers.json_load(path)
        elif Path(path).suffix == '.yaml':
            return helpers.yaml_load(path)
        elif Path(path).suffix == '.h5':
            return h5_handling.simple_load(path, return_dict=True)
        else:
            raise ValueError(f"Unknown file type: {Path(path).suffix}")

    checker = helpers.Equivalence_checker(
            kwargs_allclose={'rtol': 1e-5, 'equal_nan': True},
            assert_mode=False,
            verbose=1,
        )
    for path_test, path_true in zip(paths_rundata_test, paths_rundata_true):
        print(f"Loading run_data from {path_test}")
        data_test, data_true = load_file(path_test), load_file(path_true)
        print(f"run_data loaded. Checking equality")
        checks = checker(test=data_test, true=data_true)
        fails = [key for key, val in helpers.flatten_dict(checks).items() if val[0]==False]
        if len(fails) > 0:
            warnings.warn(f"run_data equality check failed for keys: {fails}")
        else:
            print(f"run_data equality check finished successfully")


"""
This file contains example pipelines for running the face_rhythm package. \n
In each case, a params dictionary is input, which must contain all the necessary parameters for the pipeline. \n
"""

def pipeline_basic(params):
    """
    This function runs the basic face_rhythm pipeline, similar to the
    interactive jupyter notebook: /notebooks/interactive_pipeline_basic.ipynb \n
    Note that the ROIs must be defined and saved as an ROIs.h5 file and then
    referenced in params['ROIs']['initialize']['path_file']. \n

    This pipeline performs the following steps: \n
    - Load video data \n
    - Load ROIs from file \n
    - Track points \n
    - Compute spectrograms \n
    - Perform tensor component analysis \n
    - Save results \n

    Args:
        params (dict): 
            Dictionary of parameters. See function for required fields. \n
            Also, /scripts/params_pipeline_basic.json contains an example
            parameters file. \n

    Returns:
        (dict):
            Dictionary containing the following keys: \n
            - path_config (str): Path to the config file \n
            - path_run_info (str): Path to the run_info file \n
            - directory_project (str): Path to the project directory \n
            - SEED (int): Random seed used \n
            - params (dict): Dictionary of parameters used \n
    """

    ########################################
    ## Start
    ########################################

    import face_rhythm as fr

    from pprint import pprint
    from pathlib import Path

    import cv2

    import numpy as np

    ## RESOURCE TRACKING
    # cpu_tracker = fr.helpers.CPU_Device_Checker()
    # cpu_tracker.track_utilization(
    #     interval=0.2,
    #     path_save=str(Path(directory_save) / 'cpu_tracker.csv'),
    # )
    # gpu_tracker = fr.helpers.NVIDIA_Device_Checker(device_index=0)
    # gpu_tracker.track_utilization(
    #     interval=0.2,
    #     path_save=str(Path(directory_save) / 'gpu_tracker.csv'),
    # )

    ## Initialize paths
    fr.util.system_info(verbose=True);

    SEED = _set_random_seed(
        seed=params['project']['random_seed'],
        deterministic=params['project']['random_seed'] is not None,
    )
    use_GPU = params['project']['use_GPU']


    directory_project = str(params['project']['directory_project'])
    directory_videos  = str(params['paths_videos']['directory_videos'])

    filename_videos_strMatch = params['paths_videos']['filename_videos_strMatch']

    path_config, path_run_info, directory_project = fr.project.prepare_project(
        directory_project=directory_project,
        overwrite_config=params['project']['overwrite_config'],  ## WARNING! CHECK THIS. If True, will overwrite existing config file!
        mkdir=True,
        initialize_visualization=params['project']['initialize_visualization'],
        verbose=params['project']['verbose'],
    )
    figure_saver = fr.util.Figure_Saver(
        path_config=path_config,
        **params['figure_saver'],
    )



    ########################################
    ## Prepare video data for point tracking
    ########################################

    if 'load_videos' in params['steps']:

        paths_videos = fr.helpers.find_paths(
            dir_outer=directory_videos,
            reMatch=filename_videos_strMatch,  ## string to use to search for files in directory. Uses regular expressions!
            depth=0,  ## how many folders deep to search
        )[:]

        pprint('Paths to videos:') if params['project']['verbose'] > 1 else None
        pprint(paths_videos, width=1000) if params['project']['verbose'] > 1 else None



        ## Make a `BufferedVideoReader` object for reading video file data

        videos = fr.helpers.BufferedVideoReader(
            paths_videos=paths_videos,
            **params['BufferedVideoReader']
        )

        ## Make a `Dataset_videos` object for referencing the raw video data

        data = fr.data_importing.Dataset_videos(
            bufferedVideoReader=videos,
            **params['Dataset_videos'],
        );

        ## Save the `Dataset_videos` object in the 'analysis_files' project folder

        data.save_config(path_config=path_config, overwrite=True, verbose=1)
        data.save_run_info(path_config=path_config, overwrite=True, verbose=1)
        data.save_run_data(path_config=path_config, overwrite=True, verbose=1)



    ########################################
    ## Define ROIs
    ########################################

    if 'ROIs' in params['steps']:

        ## Either select new ROIs (`select_mode='gui'`), or import existing ROIs (`path_file=path_to_ROIs.h5_file`).\
        ## Typically, you should make 1 or 2 ROIs. One for defining where the face points should be and one for cropping the frame.

        # %matplotlib notebook
        rois = fr.rois.ROIs(**params['ROIs']['initialize'])

        rois.make_points(
            rois=rois[params['ROIs']['make_points']['rois_points_idx']],
            point_spacing=params['ROIs']['make_points']['point_spacing'],
        ) if rois.point_positions is None else None

        ## Save the `ROIs` object in the 'analysis_files' project folder

        rois.save_config(path_config=path_config, overwrite=True, verbose=1)
        rois.save_run_info(path_config=path_config, overwrite=True, verbose=1)
        rois.save_run_data(path_config=path_config, overwrite=True, verbose=1)



        # ## visualize the ROIs

        # rois.plot_masks(data[0][0])



    ########################################
    # Point Tracking
    ########################################

    if 'point_tracking' in params['steps']:

        ## Prepare `PointTracker` object.\
        ## Set `visualize_video` to **`True`** to tune parameters until they look appropriate, then set to **`False`** to run the full dataset through at a much faster speed.
        ##
        ## Key parameters:
        ## - `point_spacing`: distance between points. Vary so that total number of points is appropriate.
        ## - `mesh_rigidity`: how rigid the mesh elasticity is. Vary so that points track well without drift.
        ## - `relaxation`: how quickly the points relax back to their home position. Vary so that points track well without dift.
        ## - `kwargs_method > winSize`: the spatial size of the optical flow calculation. Smaller is better but noisier, larger is less accurate but more robust to noise.
        ## - `params_outlier_handling > threshold_displacement`: point displacements above this value will result in freezing of the points.

        pt = fr.point_tracking.PointTracker(
            buffered_video_reader=videos,
            point_positions=rois.point_positions,
            rois_masks=rois[1],
            **params['PointTracker'],
        )

        ## Perform point tracking

        pt.track_points()



        ## Save the `PointTracker` object in 'analysis_files' project directory.\
        ## Using compression can reduce file sizes slightly but is very slow.

        pt.save_config(path_config=path_config, overwrite=True, verbose=1)
        pt.save_run_info(path_config=path_config, overwrite=True, verbose=2)
        pt.save_run_data(path_config=path_config, overwrite=True, use_compression=False, verbose=1)



        ## Clear some memory if needed. Optional.

        pt.cleanup()



    ########################################
    # Spectral Analysis
    ########################################

    if 'VQT' in params['steps']:

        ## Load the `PointTracker` data as a dictionary

        pt_data = fr.h5_handling.simple_load(str(Path(directory_project) / 'analysis_files' / 'PointTracker.h5'))

        ## Prepare `VQT_Analyzer` object.
        ##
        ## Key parameters:
        ## - `Q_lowF`:  Quality of the lowest frequency band of the spectrogram. Q value is number of oscillation periods.
        ## - `Q_highF`: Quality of the highest frequency band...
        ## - `F_min`: Lowest frequency band to use.
        ## - `F_max`: Highest frequency band to use.
        ## - `downsample_factor`: How much to downsample the spectrogram by in time.
        ## - `return_complex`: Whether or not to return the complex spectrogram. Generally set to False unless you want to try something fancy.

        Fs = fr.util.load_run_info_file(path_run_info)['Dataset_videos']['frame_rate']

        params['VQT_Analyzer']['params_VQT']['Fs_sample'] = Fs
        params['VQT_Analyzer']['params_VQT']['DEVICE_compute'] = fr.helpers.set_device(use_GPU=use_GPU)

        spec = fr.spectral_analysis.VQT_Analyzer(**params['VQT_Analyzer'])


        ## Look at a demo spectrogram of a single point.\
        ## Specify the point with the `idx_point` and `name_points` fields.\
        ## Note that the `pt_data['points_tracked']` dictionary holds subdictionaries withe numeric string names (ie `['0'], ['1']`) for each video.

        # demo_sepc = spec.demo_transform(
        #     points_tracked=pt_data['points_tracked'],
        #     point_positions=pt_data['point_positions'],
        #     idx_point=30,
        #     name_points='0',
        #     plot=False,
        # );

        ## Generate spectrograms

        spec.transform_all(
            points_tracked=pt_data['points_tracked'],
            point_positions=pt_data['point_positions'],
        )

        ## Save the `VQT_Analyzer` object in 'analysis_files' project directory.\
        ## Using compression can reduce file sizes slightly but is very slow.

        spec.save_config(path_config=path_config, overwrite=True, verbose=1)
        spec.save_run_info(path_config=path_config, overwrite=True, verbose=1)
        spec.save_run_data(path_config=path_config, overwrite=True, use_compression=False, verbose=1)

        ## Clear some memory if needed. Optional.

        spec.cleanup()



    ########################################
    # Decomposition
    ########################################

    if 'TCA' in params['steps']:

        ## Load the `VQT_Analyzer` data as a dictionary

        spec_data = fr.h5_handling.simple_load(str(Path(directory_project) / 'analysis_files' / 'VQT_Analyzer.h5'))

        ## Prepare `TCA` object, and then rearrange the data with the `.rearrange_data` method.
        ##
        ## Key parameters for `.rearrange_data`:
        ## - `names_dims_array`:  Enter the names of the dimensions of the spectrogram. Typically these are `'xy', 'points', 'frequency', 'time'`.
        ## - `names_dims_concat_array`: Enter any dimensions you wish to concatenate along other dimensions. Typically we wish to concatenate the `'xy'` dimension along the `'points'` dimension, so we make a list containing that pair as a tuple: `[('xy', 'points')]`.
        ## - `concat_complexDim`: If your input data are complex valued, then this can concatenate the complex dimension along another dimension.
        ## - `name_dim_dictElements`: The `data` argument is expected to be a dictionary of dictionaries of arrays, where the inner dicts are trials or videos. This is the name of what those inner dicts are. Typically `'trials'`.

        # spectrograms = spec_data['spectrograms']
        spectrograms = {key: np.abs(val) for key,val in list(spec_data['spectrograms'].items())[:]}

        tca = fr.decomposition.TCA(
            verbose=params['TCA']['verbose'],
        )

        tca.rearrange_data(
            data=spectrograms,
            **params['TCA']['rearrange_data'],
        )



        ## Fit TCA model.
        ##
        ## There are a few methods that can be used:
        ## - `'CP_NN_HALS'`: non-negative CP decomposition using the efficient HALS algorithm. This should be used in most cases.
        ## - `'CP'`: Standard CP decomposition. Use if input data are not non-negative (if you are using complex valued spectrograms or similar).
        ## - `'Randomized_CP'`: Randomized CP decomposition. Allows for large input tensors. If you are using huge tensors and you are memory constrained or want to run on a small GPU, this is your only option.
        ##
        ## If you have and want to use a CUDA compatible GPU:
        ## - Set `DEVICE` to `'cuda'`
        ## - GPU memory can be saved by setting `'init'` method to `'random'`. However, fastest convergence and highest accuracy typically come from `'init': 'svd'`.

        tca.fit(
            DEVICE=fr.helpers.set_device(use_GPU=use_GPU),
            **params['TCA']['fit'],
        )



        ## Rearrange the factors.\
        ## You can undo the concatenation that was done during `.rearrange_data`

        tca.rearrange_factors(**params['TCA']['rearrange_factors'])



        ## Save the `TCA` object in 'analysis_files' project directory.

        tca.save_config(path_config=path_config, overwrite=True, verbose=1)
        tca.save_run_info(path_config=path_config, overwrite=True, verbose=1)
        tca.save_run_data(path_config=path_config, overwrite=True, use_compression=False, verbose=1)



        ## Clear some memory if needed. Useful if you ran the fit on a GPU. Optional.

        tca._cleanup()


        # ## Plot factors

        # tca.plot_factors(
        #     figure_saver=None,
        #     show_figures=True,
        # )



        # ## Load the `TCA` data as a dictionary

        # tca_data = fr.h5_handling.simple_load(str(Path(directory_project) / 'analysis_files' / 'TCA.h5'))



    ########################################
    # Demo playback
    ########################################

    # ## Playback a video with points overlayed.\
    # ## Make sure you have a `BufferedVideoReader` object called `videos` made of your videos

    # idx_video_to_use = 0
    # idx_frames_to_use = np.arange(0,5000)

    # videos.method_getitem = 'by_video'

    # frame_visualizer = fr.visualization.FrameVisualizer(
    #     display=True,
    #     error_checking=True,
    # #     path_save=str(Path(directory_project) / 'visualizations' / 'point_tracking_demo.avi'),
    #     path_save=None,
    #     frame_height_width=videos.frame_height_width,
    #     frame_rate=240,
    #     point_sizes=3,
    #     points_colors=(0,255,255),
    #     alpha=0.3,
    # )

    # fr.visualization.play_video_with_points(
    #     bufferedVideoReader=videos[idx_video_to_use],
    #     frameVisualizer=frame_visualizer,
    #     points=list(pt_data['points_tracked'].values())[0],
    #     idx_frames=idx_frames_to_use,
    # )



    ########################################
    # Complete messages
    ########################################

    print(f'RUN COMPLETE')

    results = {
        'path_config': path_config,
        'path_run_info': path_run_info,
        'directory_project': directory_project,
        'SEED': SEED,
        'params': params,
    }

    return results


def _set_random_seed(seed=None, deterministic=False):
    """
    Set random seed for reproducibility.
    RH 2023

    Args:
        seed (int, optional):
            Random seed.
            If None, a random seed (spanning int32 integer range) is generated.
        deterministic (bool, optional):
            Whether to make packages deterministic.

    Returns:
        (int):
            seed (int):
                Random seed.
    """
    ### random seed (note that optuna requires a random seed to be set within the pipeline)
    import numpy as np
    seed = int(np.random.randint(0, 2**31 - 1, dtype=np.uint32)) if seed is None else seed

    np.random.seed(seed)
    import torch
    torch.manual_seed(seed)
    import random
    random.seed(seed)
    import cv2
    cv2.setRNGSeed(seed)

    ## Make torch deterministic
    torch.use_deterministic_algorithms(deterministic)
    ## Make cudnn deterministic
    torch.backends.cudnn.deterministic = deterministic
    torch.backends.cudnn.benchmark = not deterministic
    
    return seed


## setup.py file for roicat
from pathlib import Path
import copy
import platform

from distutils.core import setup

## Get the parent directory of this file
dir_parent = Path(__file__).parent

## Get requirements from requirements.txt
def read_requirements():
    with open(str(dir_parent / "requirements.txt"), "r") as req:
        content = req.read()  ## read the file
        requirements = content.split("\n") ## make a list of requirements split by (\n) which is the new line character

    ## Filter out any empty strings from the list
    requirements = [req for req in requirements if req]
    ## Filter out any lines starting with #
    requirements = [req for req in requirements if not req.startswith("#")]
    ## Remove any commas, quotation marks, and spaces from each requirement
    requirements = [req.replace(",", "").replace("\"", "").replace("\'", "").strip() for req in requirements]

    return requirements
deps_all = read_requirements()


## Dependencies: latest versions of requirements
### remove everything starting and after the first =,>,<,! sign
deps_names = [req.split('=')[0].split('>')[0].split('<')[0].split('!')[0] for req in deps_all]
deps_all_dict = dict(zip(deps_names, deps_all))
deps_all_latest = dict(zip(deps_names, deps_names))


# Operating system specific dependencies
# OpenCV >= 4.9 is not supported on macOS < 12
system, version_macos = platform.system(), platform.mac_ver()[0]
print(f"System: {system}")
if (system == "Darwin"):
    # Safely convert version string components to integers
    version_parts = version_macos.split('.')
    version_major_macos = int(version_parts[0])

    # Check macOS version and adjust the OpenCV version accordingly
    if (version_major_macos < 12) and ('opencv_contrib_python' in deps_all_dict):
        version_opencv_macos_sub12 = "opencv_contrib_python<=4.8.1.78"
        print(f"Detected macOS version {version_major_macos}, which is < 12. Installing an older version of OpenCV: {version_opencv_macos_sub12}")
        deps_all_dict['opencv_contrib_python'] = version_opencv_macos_sub12
        deps_all_latest['opencv_contrib_python'] = version_opencv_macos_sub12
import re
## find the numbers in the string
version_opencv = '.'.join(re.findall(r'[0-9]+', deps_all_dict['opencv_contrib_python']))
if len(version_opencv) > 0:
    version_opencv = f"<={version_opencv}"


## Make different versions of dependencies
### Also pull out the version number from the requirements (specified in deps_all_dict values).
deps_core = [deps_all_dict[dep] for dep in [
    'numpy',
    'jupyter',
    'notebook',
    'tensorly',
    'opencv_contrib_python',
    'matplotlib',
    'scikit_learn',
    'scikit_image',
    'pyyaml',
    'tqdm',
    'h5py',
    'ipywidgets',
    'Pillow',
    'eva_decord',
    'natsort',
    'pandas',
    'tables',
    'einops',
    'pytest',
    'torch',
    'torchvision',
    'torchaudio',
    'nvidia_ml_py3',
    'py_cpuinfo',
    'GPUtil',
    'psutil',
    'requests',
]]


## Make versions with cv2 headless (for servers)
deps_all_dict_cv2Headless = copy.deepcopy(deps_all_dict)
deps_all_dict_cv2Headless['opencv_contrib_python'] = 'opencv_contrib_python_headless' + version_opencv
deps_all_latest_cv2Headless = copy.deepcopy(deps_all_latest)
deps_all_latest_cv2Headless['opencv_contrib_python'] = 'opencv_contrib_python_headless'


## Get README.md
with open(str(dir_parent / "README.md"), "r") as f:
    readme = f.read()

## Get face-rhythm version number
with open(str(dir_parent / "face_rhythm" / "__init__.py"), "r") as f:
    for line in f:
        if line.startswith("__version__"):
            version = line.split("=")[1].strip().replace("\"", "").replace("\'", "")
            break


setup(
    name='face_rhythm',
    version=version,
    keywords=['neuroscience', 'neuroimaging', 'machine learning'],

    description="A pipeline for analysis of facial behavior using optical flow",
    long_description=readme,
    long_description_content_type='text/markdown',

    # The project's main homepage.
    url='https://github.com/RichieHakim/face-rhythm',

    # Author details
    author='Rich Hakim',

    # Choose your license
    license='LICENSE',

    # Supported platforms
    platforms=['Any'],

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
    ],

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    # packages=setuptools.find_packages(),
    packages=['face_rhythm'],

    install_requires=[],

    extras_require={
        'all': list(deps_all_dict.values()),
        'all_latest': list(deps_all_latest.values()),
        'all_cv2Headless': list(deps_all_dict_cv2Headless.values()),
        'all_latest_cv2Headless': list(deps_all_latest_cv2Headless.values()),
        'core': deps_core,
    },

    include_package_data=True,
)